@@ -29,16 +29,20 @@
 """Implementing communication with MySQL servers.
 """
 
+from decimal import Decimal
 from io import IOBase
+import datetime
+import getpass
 import logging
 import os
-import platform
 import socket
+import struct
 import time
+import warnings
 
 from .authentication import get_auth_plugin
 from .constants import (
-    ClientFlag, ServerCmd, ServerFlag,
+    ClientFlag, ServerCmd, ServerFlag, FieldType,
     flag_is_set, ShutdownType, NET_BUFFER_LENGTH
 )
 
@@ -51,9 +55,11 @@ from .cursor import (
     MySQLCursorBufferedNamedTuple)
 from .network import MySQLUnixSocket, MySQLTCPSocket
 from .protocol import MySQLProtocol
-from .utils import int4store, linux_distribution
+from .utils import int1store, int4store, lc_int, get_platform
 from .abstracts import MySQLConnectionAbstract
 
+logging.getLogger(__name__).addHandler(logging.NullHandler())
+
 _LOGGER = logging.getLogger(__name__)
 
 
@@ -99,6 +105,7 @@ class MySQLConnection(MySQLConnectionAbstract):
         self._auth_plugin = None
         self._krb_service_principal = None
         self._pool_config_version = None
+        self._query_attrs_supported = False
 
         self._columns_desc = []
 
@@ -113,21 +120,7 @@ class MySQLConnection(MySQLConnectionAbstract):
 
     def _add_default_conn_attrs(self):
         """Add the default connection attributes."""
-        if os.name == "nt":
-            if "64" in platform.architecture()[0]:
-                platform_arch = "x86_64"
-            elif "32" in platform.architecture()[0]:
-                platform_arch = "i386"
-            else:
-                platform_arch = platform.architecture()
-            os_ver = "Windows-{}".format(platform.win32_ver()[1])
-        else:
-            platform_arch = platform.machine()
-            if platform.system() == "Darwin":
-                os_ver = "{}-{}".format("macOS", platform.mac_ver()[0])
-            else:
-                os_ver = "-".join(linux_distribution()[0:2])
-
+        platform = get_platform()
         license_chunks = version.LICENSE.split(" ")
         if license_chunks[0] == "GPLv2":
             client_license = "GPL-2.0"
@@ -135,12 +128,15 @@ class MySQLConnection(MySQLConnectionAbstract):
             client_license = "Commercial"
         default_conn_attrs = {
             "_pid": str(os.getpid()),
-            "_platform": platform_arch,
+            "_platform": platform["arch"],
             "_source_host": socket.gethostname(),
             "_client_name": "mysql-connector-python",
             "_client_license": client_license,
-            "_client_version": ".".join([str(x) for x in version.VERSION[0:3]]),
-            "_os": os_ver}
+            "_client_version": ".".join(
+                [str(x) for x in version.VERSION[0:3]]
+            ),
+            "_os": platform["version"],
+        }
 
         self._conn_attrs.update((default_conn_attrs))
 
@@ -176,6 +172,10 @@ class MySQLConnection(MySQLConnectionAbstract):
         if handshake['capabilities'] & ClientFlag.PLUGIN_AUTH:
             self.set_client_flags([ClientFlag.PLUGIN_AUTH])
 
+        if handshake['capabilities'] & ClientFlag.CLIENT_QUERY_ATTRIBUTES:
+            self._query_attrs_supported = True
+            self.set_client_flags([ClientFlag.CLIENT_QUERY_ATTRIBUTES])
+
         self._handshake = handshake
 
     def _do_auth(self, username=None, password=None, database=None,
@@ -208,6 +208,16 @@ class MySQLConnection(MySQLConnectionAbstract):
                                        ssl_options.get('tls_versions'))
             self._ssl_active = True
 
+        _LOGGER.debug(
+                "# _do_auth(): user: %s", username)
+        _LOGGER.debug(
+                "# _do_auth(): self._auth_plugin: %s", self._auth_plugin)
+        if self._auth_plugin.startswith("authentication_oci") and not username:
+            username = getpass.getuser()
+            _LOGGER.debug(
+                "MySQL user is empty, OS user: %s will be used for "
+                "authentication_oci_client", username)
+
         packet = self._protocol.make_auth(
             handshake=self._handshake,
             username=username, password=password, database=database,
@@ -248,6 +258,13 @@ class MySQLConnection(MySQLConnectionAbstract):
             if new_auth_plugin == "authentication_ldap_sasl_client":
                 _LOGGER.debug("# auth_data: %s", auth_data)
                 response = auth.auth_response(self._krb_service_principal)
+            elif new_auth_plugin == "authentication_kerberos_client":
+                _LOGGER.debug("# auth_data: %s", auth_data)
+                response = auth.auth_response(auth_data)
+            elif new_auth_plugin == "authentication_oci_client":
+                _LOGGER.debug("# oci configuration file path: %s",
+                              self._oci_config_file)
+                response = auth.auth_response(self._oci_config_file)
             else:
                 response = auth.auth_response()
             _LOGGER.debug("# request: %s size: %s", response, len(response))
@@ -301,6 +318,51 @@ class MySQLConnection(MySQLConnectionAbstract):
                 # receive OK packet from server.
                 packet = self._socket.recv()
                 _LOGGER.debug("<< ok packet from server: %s", packet)
+            elif (
+                new_auth_plugin == "authentication_kerberos_client"
+                and packet[4] != 255
+            ):
+                rcode_size = 5  # Reader size for the response status code
+                _LOGGER.debug("# Continue with GSSAPI authentication")
+                _LOGGER.debug("# Response header: %s", packet[:rcode_size + 1])
+                _LOGGER.debug("# Response size: %s", len(packet))
+                _LOGGER.debug("# Negotiate a service request")
+                complete = False
+                tries = 0
+
+                while not complete and tries < 5:
+                    _LOGGER.debug(
+                        "%s Attempt %s %s", "-" * 20, tries + 1, "-" * 20
+                    )
+                    _LOGGER.debug("<< Server response: %s", packet)
+                    _LOGGER.debug(
+                        "# Response code: %s", packet[:rcode_size + 1]
+                    )
+                    token, complete = auth.auth_continue(packet[rcode_size:])
+                    if token:
+                        self._socket.send(token)
+                    if complete:
+                        break
+                    packet = self._socket.recv()
+
+                    _LOGGER.debug(">> Response to server: %s", token)
+                    tries += 1
+
+                if not complete:
+                    raise errors.InterfaceError(
+                        "Unable to fulfill server request after {} attempts. "
+                        "Last server response: {}".format(tries, packet)
+                    )
+
+                _LOGGER.debug(
+                    "Last response from server: %s length: %d",
+                    packet,
+                    len(packet),
+                )
+
+                # Receive OK packet from server.
+                packet = self._socket.recv()
+                _LOGGER.debug("<< Ok packet from server: %s", packet)
 
         if packet[4] == 1:
             auth_data = self._protocol.parse_auth_more_data(packet)
@@ -344,6 +406,20 @@ class MySQLConnection(MySQLConnectionAbstract):
 
         Raises on errors.
         """
+        if any((self._password1, self._password2, self._password3)):
+            raise errors.NotSupportedError(
+                "The Multi Factor Authentication is not supported by the "
+                "pure Python implementation"
+            )
+        if self._auth_plugin == "authentication_kerberos_client":
+            if os.name == "nt":
+                raise errors.ProgrammingError(
+                    "The Kerberos authentication is not available on Windows"
+                )
+            if not self._user:
+                cls = get_auth_plugin(self._auth_plugin)
+                self._user = cls.get_user_from_credentials()
+
         self._protocol = MySQLProtocol()
         self._socket = self._get_connection()
         try:
@@ -362,6 +438,21 @@ class MySQLConnection(MySQLConnectionAbstract):
             self.close()
             raise
 
+        if (
+            not self._ssl_disabled
+            and hasattr(self._socket.sock, "version")
+            and callable(self._socket.sock.version)
+        ):
+            # Raise a deprecation warning if TLSv1 or TLSv1.1 is being used
+            tls_version = self._socket.sock.version()
+            if tls_version in ("TLSv1", "TLSv1.1"):
+                warn_msg = (
+                    f"This connection is using {tls_version} which is now "
+                    "deprecated and will be removed in a future release of "
+                    "MySQL Connector/Python"
+                )
+                warnings.warn(warn_msg, DeprecationWarning)
+
     def shutdown(self):
         """Shut down connection to MySQL Server.
         """
@@ -680,8 +771,85 @@ class MySQLConnection(MySQLConnectionAbstract):
 
         Returns a tuple()
         """
-        if not isinstance(query, bytes):
-            query = query.encode('utf-8')
+        if not isinstance(query, bytearray):
+            if isinstance(query, str):
+                query = query.encode('utf-8')
+            query = bytearray(query)
+        # Prepare query attrs
+        charset = self.charset if self.charset != "utf8mb4" else "utf8"
+        packet = bytearray()
+        if not self._query_attrs_supported and self._query_attrs:
+            warnings.warn(
+                "This version of the server does not support Query Attributes",
+                category=Warning)
+        if self._client_flags & ClientFlag.CLIENT_QUERY_ATTRIBUTES:
+            names = []
+            types = []
+            values = []
+            null_bitmap = [0] * ((len(self._query_attrs) + 7) // 8)
+            for pos, attr_tuple in enumerate(self._query_attrs):
+                value = attr_tuple[1]
+                flags = 0
+                if value is None:
+                    null_bitmap[(pos // 8)] |= 1 << (pos % 8)
+                    types.append(int1store(FieldType.NULL) +
+                                 int1store(flags))
+                    continue
+                elif isinstance(value, int):
+                    (packed, field_type,
+                     flags) = self._protocol._prepare_binary_integer(value)
+                    values.append(packed)
+                elif isinstance(value, str):
+                    value = value.encode(charset)
+                    values.append(lc_int(len(value)) + value)
+                    field_type = FieldType.VARCHAR
+                elif isinstance(value, bytes):
+                    values.append(lc_int(len(value)) + value)
+                    field_type = FieldType.BLOB
+                elif isinstance(value, Decimal):
+                    values.append(
+                        lc_int(len(str(value).encode(
+                            charset))) + str(value).encode(charset))
+                    field_type = FieldType.DECIMAL
+                elif isinstance(value, float):
+                    values.append(struct.pack('<d', value))
+                    field_type = FieldType.DOUBLE
+                elif isinstance(value, (datetime.datetime, datetime.date)):
+                    (packed, field_type) = \
+                       self._protocol._prepare_binary_timestamp(value)
+                    values.append(packed)
+                elif isinstance(value, (datetime.timedelta, datetime.time)):
+                    (packed, field_type) = \
+                       self._protocol._prepare_binary_time(value)
+                    values.append(packed)
+                else:
+                    raise errors.ProgrammingError(
+                        "MySQL binary protocol can not handle "
+                        "'{classname}' objects".format(
+                            classname=value.__class__.__name__))
+                types.append(int1store(field_type) +
+                             int1store(flags))
+                name = attr_tuple[0].encode(charset)
+                names.append(lc_int(len(name)) + name)
+
+            # int<lenenc>    parameter_count    Number of parameters
+            packet.extend(lc_int(len(self._query_attrs)))
+            # int<lenenc>    parameter_set_count    Number of parameter sets.
+            # Currently always 1
+            packet.extend(lc_int(1))
+            if values:
+                packet.extend(
+                    b''.join([struct.pack('B', bit) for bit in null_bitmap]) +
+                    int1store(1))
+                for _type, name in zip(types, names):
+                    packet.extend(_type)
+                    packet.extend(name)
+
+                for value in values:
+                    packet.extend(value)
+
+        packet.extend(query)
+        query = bytes(packet)
         try:
             result = self._handle_result(self._send_cmd(ServerCmd.QUERY, query))
         except errors.ProgrammingError as err:
@@ -714,13 +882,23 @@ class MySQLConnection(MySQLConnectionAbstract):
 
         Returns a generator.
         """
+        packet = bytearray()
         if not isinstance(statements, bytearray):
             if isinstance(statements, str):
                 statements = statements.encode('utf8')
             statements = bytearray(statements)
 
+        if self._client_flags & ClientFlag.CLIENT_QUERY_ATTRIBUTES:
+            # int<lenenc>    parameter_count    Number of parameters
+            packet.extend(lc_int(0))
+            # int<lenenc>    parameter_set_count    Number of parameter sets.
+            # Currently always 1
+            packet.extend(lc_int(1))
+
+        packet.extend(statements)
+        query = bytes(packet)
         # Handle the first query result
-        yield self._handle_result(self._send_cmd(ServerCmd.QUERY, statements))
+        yield self._handle_result(self._send_cmd(ServerCmd.QUERY, query))
 
         # Handle next results, if any
         while self._have_next_result:
@@ -829,7 +1007,8 @@ class MySQLConnection(MySQLConnectionAbstract):
         return self._handle_ok(self._send_cmd(ServerCmd.PING))
 
     def cmd_change_user(self, username='', password='', database='',
-                        charset=45):
+                        charset=45, password1='', password2='', password3='',
+                        oci_config_file=''):
         """Change the current logged in user
 
         This method allows to change the current logged in user information.
@@ -839,6 +1018,11 @@ class MySQLConnection(MySQLConnectionAbstract):
         """
         self.handle_unread_result()
 
+        if any((password1, password2, password3)):
+            raise errors.NotSupportedError(
+                "The Multi Factor Authentication is not supported by the "
+                "pure Python implementation"
+            )
         if self._compress:
             raise errors.NotSupportedError("Change user is not supported with "
                                            "compression.")
@@ -851,6 +1035,9 @@ class MySQLConnection(MySQLConnectionAbstract):
             conn_attrs=self._conn_attrs)
         self._socket.send(packet, 0, 0)
 
+        if oci_config_file:
+            self._oci_config_file = oci_config_file
+
         ok_packet = self._auth_switch_request(username, password)
 
         try:
@@ -920,7 +1107,10 @@ class MySQLConnection(MySQLConnectionAbstract):
             self.cmd_reset_connection()
         except errors.NotSupportedError:
             self.cmd_change_user(self._user, self._password,
-                                 self._database, self._charset_id)
+                                 self._database, self._charset_id,
+                                 self._password1, self._password2,
+                                 self._password3,
+                                 self._oci_config_file)
 
         cur = self.cursor()
         if user_variables:
@@ -1191,10 +1381,20 @@ class MySQLConnection(MySQLConnectionAbstract):
                     self.cmd_stmt_send_long_data(statement_id, param_id,
                                                  data[param_id])
                     long_data_used[param_id] = (binary,)
-
-        execute_packet = self._protocol.make_stmt_execute(
-            statement_id, data, tuple(parameters), flags,
-            long_data_used, self.charset)
+        if not self._query_attrs_supported and self._query_attrs:
+            warnings.warn(
+                "This version of the server does not support Query Attributes",
+                category=Warning)
+        if self._client_flags & ClientFlag.CLIENT_QUERY_ATTRIBUTES:
+            execute_packet = self._protocol.make_stmt_execute(
+                statement_id, data, tuple(parameters), flags,
+                long_data_used, self.charset, self._query_attrs,
+                self._converter_str_fallback)
+        else:
+            execute_packet = self._protocol.make_stmt_execute(
+                statement_id, data, tuple(parameters), flags,
+                long_data_used, self.charset,
+                converter_str_fallback=self._converter_str_fallback)
         packet = self._send_cmd(ServerCmd.STMT_EXECUTE, packet=execute_packet)
         result = self._handle_binary_result(packet)
         return result
